"""Assembly results.

All the assembly tools must have a common set of results.
"""

from __future__ import annotations

from pathlib import Path
from typing import final

from slurmbench.prelude.topic import results as core

import benchmark.topics.assembly.visitor as asm_visitor


@final
class FastaGZ(core.Original):
    """FASTA gunzip result."""

    FASTA_GZ_NAME = Path("assembly.fasta.gz")

    def fasta_gz(self, sample_dirname: str | Path) -> Path:
        """Get assembly FASTA file."""
        return self._exp_fs_manager.sample_dir(sample_dirname) / self.FASTA_GZ_NAME


@final
class AsmGraphGZ(core.Original):
    """Assembly graph (GFA) gunzip result."""

    ASSEMBLY_GFA_GZ_NAME = Path("assembly.gfa.gz")

    def gfa_gz(self, sample_dirname: str | Path) -> Path:
        """Get assembly GFA file."""
        return (
            self._exp_fs_manager.sample_dir(sample_dirname) / self.ASSEMBLY_GFA_GZ_NAME
        )


@final
class FastaGZVisitor(core.OriginalVisitor):
    """FastaGZ result visitor."""

    @classmethod
    def result_builder(cls) -> type[FastaGZ]:
        """Get result builder."""
        return FastaGZ

    @classmethod
    def result_builder_from_tool(
        cls,
        tool: asm_visitor.Tools,
    ) -> core.Error | type[FastaGZ]:
        """Visit assembly FASTA tool result."""
        match tool:
            case asm_visitor.Tools.UNICYCLER:
                return cls.result_builder()

            # FEATURE: Add SKESA and GFA_CONNECTOR
            # case asm_visitor.Tools.SKESA:
            #     return cls.result_builder()
            # case asm_visitor.Tools.GFA_CONNECTOR:
            #     return cls.result_builder()


@final
class AsmGraphGZVisitor(core.OriginalVisitor):
    """Assembly graph (GFA) result visitor."""

    @classmethod
    def result_builder(cls) -> type[AsmGraphGZ]:
        """Get result builder."""
        return AsmGraphGZ

    @classmethod
    def result_builder_from_tool(
        cls,
        tool: asm_visitor.Tools,
    ) -> core.Error | type[AsmGraphGZ]:
        """Visit assembly graph (GFA) tool result."""
        match tool:
            case asm_visitor.Tools.UNICYCLER:
                return cls.result_builder()

            # FEATURE: Add SKESA and GFA_CONNECTOR
            # case asm_visitor.Tools.SKESA:
            #     return core.Error(
            #         f"{asm_visitor.Tools.SKESA} tool"
            #         " does not provide a GFA file"
            #         f" but {asm_visitor.Tools.GFA_CONNECTOR} does",
            #     )
            # case asm_visitor.Tools.GFA_CONNECTOR:
            #     return cls.result_builder()



"""Init experiment logics."""

from __future__ import annotations

import logging

# import pbfbench.abc.tool.connector as abc_tool_connector
# import pbfbench.abc.topic.results as abc_topic_res
# import pbfbench.experiment.file_system as exp_fs
# import pbfbench.experiment.iter as exp_iter
# import pbfbench.samples.file_system as smp_fs
# import pbfbench.topics.binning.pangebin_once.config as pangebin_once_cfg
# import pbfbench.topics.classification.pbf_input.results as class_pbf_in_res
# import pbfbench.topics.classification.plasgraph2.plasbin_flow as plasgraph2_pbf
# import pbfbench.topics.classification.platon.convert as platon_pbf
# import pbfbench.topics.classification.visitor as class_visitor

_LOGGER = logging.getLogger(__name__)


# TODO [2025-09-19 18:41] Try to generalize
# def init(
#     data_exp_fs_manager: exp_fs.DataManager,
#     exp_config: exp_cfg.WithArguments,
#     tool_connector: abc_tool_connector.WithArguments,
# ) -> InitStats:
#     """Init pangebin-once."""
#     init_stats = InitStats.new(data_exp_fs_manager)

#     #
#     # Seeds
#     #
#     # TODO verify config exists (use already existing functions) [Not urgent]
#     # REFACTOR generalize this but e.g. plasmidness needs also the GFA tool provider
#     # REFACTOR to get the subconnector, use instead class attribute for Connector?
#     # * Connector must implement abc classmethod to_arg_paths -> Iterable[ArgPath]
#     # IDEA InitConnector with inputExpConfig type and FormattedVisitor
#     # FormattedVisitor links to FormattedResult type and convert functions
#     _tool_cfg: abc_tool_connector.WithArguments = exp_config.tool_configs()
#     seeds_arg = _tool_cfg.arguments()[pangebin_once_cfg.Names.SEEDS]
#     seeds_tool = class_visitor.Tools(seeds_arg.tool())
#     seeds_in_data_exp_fs_manager = exp_fs.DataManager(
#         data_exp_fs_manager.root_dir(),
#         seeds_tool.to_description(),
#         seeds_arg.exp_name(),
#     )
#     match seeds_tool:
#         case class_visitor.Tools.PLATON:
#             convert_function = platon_pbf.convert
#         # FIXME force match cover with return
#     samples_to_format_the_seeds = _get_samples_to_format_the_inputs(
#         data_exp_fs_manager,
#         class_pbf_in_res.Seeds(seeds_in_data_exp_fs_manager),
#         init_stats,
#     )

#     for sample in samples_to_format_the_seeds:
#         # FIXME not good usage of init stats, think in another way
#         init_stats.add_samples_to_format_the_inputs(1)
#         convert_function(
#             seeds_in_data_exp_fs_manager,
#             sample.item(),
#         )

#     #
#     # Plasmidness
#     #
#     # REFACTOR same refactor comment as above

#     class_arg = _tool_cfg.arguments()[pangebin_once_cfg.Names.PLASMIDNESS]
#     class_tool = class_visitor.Tools(class_arg.tool())
#     class_in_data_exp_fs_manager = exp_fs.DataManager(
#         data_exp_fs_manager.root_dir(),
#         class_tool.to_description(),
#         class_arg.exp_name(),
#     )
#     match class_tool:
#         case class_visitor.Tools.PLASCLASS:
#             raise NotImplementedError  # TODO implement for plasclass
#         case class_visitor.Tools.PLASGRAPH2:
#             convert_function = plasgraph2_pbf.convert
#         # REFACTOR force match cover with return

#     samples_to_format_the_plm = _get_samples_to_format_the_inputs(
#         data_exp_fs_manager,
#         class_pbf_in_res.Plasmidness(class_in_data_exp_fs_manager),
#         init_stats,
#     )
#     for sample in samples_to_format_the_plm:
#         # REFACTOR (1) here we simulate that Visitor
#         convert_function(
#             class_in_data_exp_fs_manager,
#             sample.item(),
#         )

#     return init_stats


# TODO [2025-09-19 18:31:24] invent a init object with logics (functions and classes)
# def _get_samples_to_format_the_inputs(
#     samples_to_run: list[smp_fs.RowNumberedItem],
#     formatted_result_builder: abc_topic_res.Formatted,
#     init_stats: InitStats,  # REFACTOR try to not complicate with init stats
# ) -> list[smp_fs.RowNumberedItem]:
#     """Get samples to run."""
#     # REFACTOR should be general not specific to pangebin_once tool
#     samples_to_fmt_the_inputs = list(
#         exp_iter.samples_to_format_result(
#             formatted_result_builder,
#             samples_to_run,
#         ),
#     )  # FIXME should not be recomputed but according to sample to run list
#     init_stats.add_samples_to_format_the_inputs(len(samples_to_fmt_the_inputs))

#     _LOGGER.info("Number of samples to run: %d", len(samples_to_fmt_the_inputs))
#     # REFACTOR also give the list of already formatted input?
#     return samples_to_fmt_the_inputs



"""plASgraph2 description."""

from slurmbench.prelude.tool import description as core

import benchmark.topics.classification.description as class_desc

DESCRIPTION = core.Description(
    "PLASGRAPH2",
    "plasgraph2",
    class_desc.DESCRIPTION,
)



"""Platon description module."""

from slurmbench.prelude.tool import description as core

import benchmark.topics.classification.description as class_desc

DESCRIPTION = core.Description("PLATON", "platon", class_desc.DESCRIPTION)



"""plASgraph2 application module."""

# Due to typer usage:

from __future__ import annotations

from slurmbench.prelude.tool import app as core

from . import connector

APP = core.new(connector.Connector)



"""plASgraph2 Bash script logics."""

from collections.abc import Iterator
from pathlib import Path
from typing import final

from slurmbench.prelude.tool import bash as core

import benchmark.topics.assembly.results as asm_res
import benchmark.topics.classification.plasgraph2.results as plasgraphtwo_res


@final
class GFAInputLinesBuilder(core.Argument[asm_res.AsmGraphGZ]):
    """GFA input bash lines builder."""

    GFA_GZ_VAR = core.BashVar("GFA")

    OUTFILE_VAR = core.BashVar("OUTFILE")

    def __gfa_gz_file(self) -> Path:
        """Return a gzipped GFA path with sample name is a sh variable."""
        return self._input_result.gfa_gz(
            self._input_data_smp_sh_fs_manager.sample_dir().name,
        )

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield self.GFA_GZ_VAR.set_path(self.__gfa_gz_file())
        yield self.OUTFILE_VAR.set_path(
            plasgraphtwo_res.PlasmidProbabilities(self._work_exp_fs_manager).csv(
                self._work_smp_sh_fs_manager.sample_dir().name,
            ),
        )



"""Tool description."""

from slurmbench.prelude.tool import description as core

import benchmark.topics.binning.description as desc

DESCRIPTION = core.Description(
    "PANGEBIN_ONCE",
    "pangebin-once",
    desc.DESCRIPTION,
)



"""Assembly topic visitor."""

from typing import final

from slurmbench.prelude.topic import visitor as core

# from .gfa_connector import description as gfa_connector_desc
# from .skesa import description as skesa_desc
from .unicycler import description as unicycler_desc


@final
class Tools(core.Tools):
    """Assembly tools descriptions."""

    UNICYCLER = unicycler_desc.DESCRIPTION.name()

    # SKESA = skesa_desc.DESCRIPTION.name()
    # GFA_CONNECTOR = gfa_connector_desc.DESCRIPTION.name()

    def to_description(self) -> core.Description:
        """Get tool description."""
        match self:
            case Tools.UNICYCLER:
                return unicycler_desc.DESCRIPTION
            # FEATURE: Add SKESA and GFA_CONNECTOR
            # case Tools.SKESA:
            #     return skesa_desc.DESCRIPTION
            # case Tools.GFA_CONNECTOR:
            #     return gfa_connector_desc.DESCRIPTION



"""Concrete topic application module."""

# Due to typer usage:

from __future__ import annotations

from slurmbench.prelude.topic import app as core

import benchmark.topics.assembly.description as assembly_desc
import benchmark.topics.assembly.unicycler.app as unicycler_app

from . import visitor

APP = core.Topic.new(assembly_desc.DESCRIPTION, visitor.Tools, [unicycler_app.APP])



"""Classification topic application module."""

# Due to typer usage:

from __future__ import annotations

from slurmbench.prelude.topic import app as core

from . import description as desc
from . import visitor
from .plasclass import app as plasclass_app
from .plasgraph2 import app as plasgraphtwo_app
from .platon import app as platon_app

APP = core.Topic.new(
    desc.DESCRIPTION,
    visitor.Tools,
    [
        plasclass_app.APP,
        plasgraphtwo_app.APP,
        platon_app.APP,
    ],
)



"""Concrete tool connector module."""

from __future__ import annotations

from typing import final

from slurmbench.prelude.tool import connector as core

import benchmark.topics.assembly.results as asm_res
import benchmark.topics.assembly.visitor as asm_visitor
import benchmark.topics.classification.visitor as class_visitor
from benchmark.topics.binning.plasbin_flow.format.classification import (
    results as fmt_class_res,
)
from benchmark.topics.binning.plasbin_flow.format.classification import (
    visitor as fmt_class_visitor,
)

from . import description as desc
from . import shell as sh


@final
class Names(core.Names):
    """Argument names."""

    GFA = "GFA"
    SEEDS = "SEEDS"
    PLASMIDNESS = "PLASMIDNESS"

    def topic_tools(self) -> type[core.Tools]:
        """Get topic tools."""
        match self:
            case Names.GFA:
                return asm_visitor.Tools
            case Names.SEEDS:
                return class_visitor.Tools
            case Names.PLASMIDNESS:
                return class_visitor.Tools


@final
class GFAArg(core.Arg[Names, asm_visitor.Tools, asm_res.AsmGraphGZ]):
    """GFA argument."""

    @classmethod
    def name(cls) -> Names:
        """Get name."""
        return Names.GFA

    @classmethod
    def tools_type(cls) -> type[asm_visitor.Tools]:
        """Get tools type."""
        return asm_visitor.Tools

    @classmethod
    def result_visitor(cls) -> type[asm_res.AsmGraphGZVisitor]:
        """Get result visitor."""
        return asm_res.AsmGraphGZVisitor

    @classmethod
    def sh_lines_builder_type(cls) -> type[sh.GFAInputLinesBuilder]:
        """Get shell lines builder type."""
        return sh.GFAInputLinesBuilder


@final
class SeedsArg(core.Arg[Names, class_visitor.Tools, fmt_class_res.Seeds]):
    """Seeds argument."""

    @classmethod
    def name(cls) -> Names:
        """Get name."""
        return Names.SEEDS

    @classmethod
    def tools_type(cls) -> type[class_visitor.Tools]:
        """Get tools type."""
        return class_visitor.Tools

    @classmethod
    def result_visitor(cls) -> type[fmt_class_visitor.SeedsVisitor]:
        """Get result visitor."""
        return fmt_class_visitor.SeedsVisitor

    @classmethod
    def sh_lines_builder_type(cls) -> type[sh.SeedsInputLinesBuilder]:
        """Get shell lines builder type."""
        return sh.SeedsInputLinesBuilder


@final
class PlasmidnessArg(core.Arg[Names, class_visitor.Tools, fmt_class_res.Plasmidness]):
    """Plasmidness argument."""

    @classmethod
    def name(cls) -> Names:
        """Get name."""
        return Names.PLASMIDNESS

    @classmethod
    def tools_type(cls) -> type[class_visitor.Tools]:
        """Get tools type."""
        return class_visitor.Tools

    @classmethod
    def result_visitor(cls) -> type[fmt_class_visitor.PlasmidnessVisitor]:
        """Get result visitor."""
        return fmt_class_visitor.PlasmidnessVisitor

    @classmethod
    def sh_lines_builder_type(
        cls,
    ) -> type[sh.PlasmidnessInputLinesBuilder]:
        """Get shell lines builder type."""
        return sh.PlasmidnessInputLinesBuilder


@final
class Arguments(core.Arguments[Names]):
    """Concrete tool arguments."""

    @classmethod
    def arg_types(cls) -> list[type[core.Arg]]:
        """Get list of arg types."""
        return [
            GFAArg,
            SeedsArg,
            PlasmidnessArg,
        ]


@final
class Connector(core.WithArguments[Names]):
    """Concrete tool connector."""

    @classmethod
    def description(cls) -> core.Description:
        """Get description."""
        return desc.DESCRIPTION

    @classmethod
    def arguments_type(cls) -> type[Arguments]:
        """Get arguments type."""
        return Arguments



"""Platon Bash script logics."""

from collections.abc import Iterator
from pathlib import Path
from typing import final

from slurmbench.prelude.tool import bash as core

import benchmark.topics.assembly.results as asm_res
import benchmark.topics.classification.plasclass.results as plasclass_res


@final
class FastaInputLinesBuilder(core.Argument[asm_res.FastaGZ]):
    """Fasta input bash lines builder."""

    FASTA_GZ_VAR = core.BashVar("FASTA_GZ")

    FASTA_VAR = core.BashVar("FASTA")
    OUTFILE_VAR = core.BashVar("OUTFILE")

    def __fasta_gz_file(self) -> Path:
        """Return a gzipped FASTA path with sample name is a sh variable."""
        return self._input_result.fasta_gz(
            self._input_data_smp_sh_fs_manager.sample_dir().name,
        )

    def __fasta_tmp_file(self) -> Path:
        """Return a tmp FASTA path with sample name is a sh variable."""
        return (
            self._work_smp_sh_fs_manager.sample_dir()
            / self._input_result.FASTA_GZ_NAME.with_suffix("")
        )

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield self.FASTA_GZ_VAR.set_path(self.__fasta_gz_file())
        yield self.FASTA_VAR.set_path(self.__fasta_tmp_file())
        yield self.OUTFILE_VAR.set_path(
            plasclass_res.PlasmidProbabilities(self._work_exp_fs_manager).tsv(
                self._work_smp_sh_fs_manager.sample_dir().name,
            ),
        )



"""Topic application module."""

# Due to typer usage:

from __future__ import annotations

from slurmbench.topic import app as core

from . import description as desc
from . import visitor
from .pangebin_once import app as pg_once_app

APP = core.Topic.new(
    desc.DESCRIPTION,
    visitor.Tools,
    [pg_once_app.APP],
)



"""SKESA assembler items."""

from slurmbench.prelude.tool import description as core

import benchmark.topics.assembly.description as asm_desc

DESCRIPTION = core.Description("SKESA", "skesa", asm_desc.DESCRIPTION)



"""Connector for Unicycler."""

# Due to typer usage:

from __future__ import annotations

from typing import final

from slurmbench.prelude.tool import connector as core

from . import bash
from . import description as desc


@final
class Connector(core.OnlyOptions):
    """Unicycler connector."""

    @classmethod
    def commands_type(cls) -> type[bash.OnlyOptions]:
        """Get custom commands type."""
        return bash.OnlyOptions

    @classmethod
    def description(cls) -> core.Description:
        """Get description."""
        return desc.DESCRIPTION



"""Plasmidness topic visitor."""

from typing import final

from slurmbench.prelude.topic import visitor as core

from .plasclass import description as plasclass_desc
from .plasgraph2 import description as plasgraph2_desc
from .platon import description as platon_desc


@final
class Tools(core.Tools):
    """Plasmidness tools descriptions."""

    PLASCLASS = plasclass_desc.DESCRIPTION.name()
    PLASGRAPH2 = plasgraph2_desc.DESCRIPTION.name()
    PLATON = platon_desc.DESCRIPTION.name()

    def to_description(self) -> core.Description:
        """Get tool description."""
        match self:
            case Tools.PLASCLASS:
                return plasclass_desc.DESCRIPTION
            case Tools.PLASGRAPH2:
                return plasgraph2_desc.DESCRIPTION
            case Tools.PLATON:
                return platon_desc.DESCRIPTION



"""Platon Bash script logics."""

from collections.abc import Iterator
from pathlib import Path
from typing import final

from slurmbench.prelude.tool import bash as core

import benchmark.topics.assembly.results as asm_res


@final
class GenomeInputLinesBuilder(core.Argument[asm_res.FastaGZ]):
    """Genome input bash lines builder."""

    GENOME_VAR = core.BashVar("GENOME")
    FASTA_GZ_VAR = core.BashVar("FASTA_GZ")

    def __fasta_gz_file(self) -> Path:
        """Return a gzipped FASTA path with sample name is a sh variable."""
        return self._input_result.fasta_gz(
            self._work_smp_sh_fs_manager.sample_dir().name,
        )

    def __fasta_tmp_file(self) -> Path:
        """Return a tmp FASTA path with sample name is a sh variable."""
        return (
            self._work_smp_sh_fs_manager.sample_dir()
            / self._input_result.FASTA_GZ_NAME.with_suffix("")
        )

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield self.FASTA_GZ_VAR.set_path(self.__fasta_gz_file())
        yield self.GENOME_VAR.set_path(self.__fasta_tmp_file())



"""Samples module."""

from enum import StrEnum

from slurmbench.prelude import samples as core


class TSVHeader(StrEnum):
    """Custom samples file header."""

    SAMPLE_UID = core.TSVHeader.SAMPLE_UID
    SHORT_READS = "short_reads"



"""SKESA assembler items."""

from slurmbench.prelude.tool import description as core

import benchmark.topics.assembly.description as asm_desc

DESCRIPTION = core.Description(
    "GFA_CONNECTOR",
    "gfa-connector",
    asm_desc.DESCRIPTION,
)



"""Platon Bash script logics."""

from collections.abc import Iterator
from typing import final

from slurmbench.prelude.tool import bash as core

from benchmark import samples as smp


@final
class OnlyOptions(core.OnlyOptions):
    """Unicycler custom OnlyOption bash command class."""

    SRR_ID_VAR = core.BashVar("SRR_ID")

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield self.SRR_ID_VAR.set(
            core.get_sample_attribute(
                self.data_exp_fs_manager().samples_tsv(),
                smp.TSVHeader.SHORT_READS,
            ),
        )



"""Tool results."""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from slurmbench.tool import results as core

import benchmark.topics.assembly.results as asm_res

if TYPE_CHECKING:
    from pathlib import Path


@final
class PlasmidStats(core.Original):
    """Plasmid stats result."""

    # TSV name: "assembly.tsv"
    # * no `.fasta` because Platon removes the extension
    # * no `.gz` because Platon takes an not-compressed FASTA file
    TSV_NAME = asm_res.FastaGZ.FASTA_GZ_NAME.with_suffix("").with_suffix(".tsv")

    def tsv(self, sample_dirname: str | Path) -> Path:
        """Get TSV file."""
        return self._exp_fs_manager.sample_dir(sample_dirname) / self.TSV_NAME



"""PlasBin-flow classification result formatting module."""

from typing import final

from slurmbench.prelude.topic import results as core

import benchmark.topics.classification.visitor as class_visitor

from . import platon, results


@final
class PlasmidnessVisitor(
    core.FormattedVisitor[class_visitor.Tools, results.Plasmidness],
):
    """Plasmidness result visitor."""

    @classmethod
    def convert_fn(
        cls,
        tool: class_visitor.Tools,
    ) -> core.ConvertFn[results.Plasmidness] | core.Error:
        """Get convert function."""

        def _err(tool: class_visitor.Tools) -> core.Error:
            return core.Error(
                "Function to convert classification result to plasmidness result"
                f" is not implemented for `{tool}` ",
            )

        match tool:
            case class_visitor.Tools.PLATON:
                return platon.plasmidness
            case class_visitor.Tools.PLASCLASS:
                return _err(tool)  # FEATURE PlasClass plasmidness convert
            case class_visitor.Tools.PLASGRAPH2:
                return _err(tool)  # FEATURE PlasGraph2 plasmidness convert

    @classmethod
    def result_builder(cls) -> type[results.Plasmidness]:
        """Get result builder."""
        return results.Plasmidness


@final
class SeedsVisitor(core.FormattedVisitor[class_visitor.Tools, results.Seeds]):
    """Seeds result visitor."""

    @classmethod
    def convert_fn(
        cls,
        tool: class_visitor.Tools,
    ) -> core.ConvertFn[results.Seeds] | core.Error:
        """Get convert function."""

        def _err(tool: class_visitor.Tools) -> core.Error:
            return core.Error(
                "Function to convert classification result to plasmidness result"
                f" is not implemented for `{tool}` ",
            )

        match tool:
            case class_visitor.Tools.PLATON:
                return platon.seeds
            case class_visitor.Tools.PLASCLASS:
                return _err(tool)  # FEATURE PlasClass seeds convert
            case class_visitor.Tools.PLASGRAPH2:
                return _err(tool)  # FEATURE PlasGraph2 seeds convert

    @classmethod
    def result_builder(cls) -> type[results.Seeds]:
        """Get result builder."""
        return results.Seeds



"""PlasBin-flow classification result formatting module."""

from pathlib import Path
from typing import final

from slurmbench.prelude.tool import results as core


@final
class Plasmidness(core.Formatted):
    """Plasmidness PlasBin-flow formatted result."""

    TSV_NAME = Path("pbf_plasmidness.tsv")

    def tsv(self, sample_dirname: str | Path) -> Path:
        """Get plasmidness TSV file."""
        return self._exp_fs_manager.sample_dir(sample_dirname) / self.TSV_NAME

    def check(self, sample_item: core.Sample) -> core.SampleStatus:
        """Check input(s)."""
        if self.tsv(sample_item.uid()).exists():
            return core.SampleSuccess.OK
        return core.SampleError.NOT_RUN


@final
class Seeds(core.Formatted):
    """Seeds PlasBin-flow formatted result."""

    TSV_NAME = Path("pbf_seeds.tsv")

    def tsv(self, sample_dirname: str | Path) -> Path:
        """Get seeds TSV file."""
        return self._exp_fs_manager.sample_dir(sample_dirname) / self.TSV_NAME

    def check(self, sample_item: core.Sample) -> core.SampleStatus:
        """Check input(s)."""
        if self.tsv(sample_item.uid()).exists():
            return core.SampleSuccess.OK
        return core.SampleError.NOT_RUN



"""PlasBin-flow plasmidness input operations logics."""

import gzip
from pathlib import Path

import benchmark.topics.assembly.visitor as asm_visitor


def parse_gfa(gfa_gz: Path, assembler: asm_visitor.Tools) -> dict:
    """Parse GFA file and return contigs dictionary."""
    contigs_dict = {}
    with gzip.open(gfa_gz, "r") as f:
        for line in f:
            gfa_line_array = line.decode()
            if gfa_line_array[0] == "S":
                match assembler:
                    # FEATURE: Add SKESA and GFA_CONNECTOR
                    # case asm_visitor.Tools.GFA_CONNECTOR:
                    #     details = gfa_line_array[:-1].split("\t")
                    #     contig_id = details[1]
                    #     contig_name = (
                    #         details[0] + details[1] + "_" + "_".join(details[3:])
                    #     )
                    #     # FIXME [ANIKET] why?
                    #     # Because contig name in FASTA and GFA differ
                    #     # GFA contig names have ':' in them to figure out a GFA segment
                    #     # is a subsequence of a FASTA contig
                    #     # FIXME [ANIKET] why renamed contig name for SKESA (_ instead of :)
                    case asm_visitor.Tools.UNICYCLER:
                        details = gfa_line_array.split("\t")
                        contig_id = details[1]
                        contig_name = str(contig_id)
                    case _:
                        _err_msg = f"Unknown assembler {assembler}"
                        raise ValueError(_err_msg)

                contigs_dict[contig_id] = {
                    "Prob_Chromosome": 0.5,
                    "Prob_Plasmid": 0.5,
                    "Prediction": "Chromosome",
                    "Contig_name": contig_name,
                    "Contig_length": len(details[2]),
                }
    return contigs_dict



"""Topics module."""

from typing import final

from slurmbench.prelude.topics import visitor as core

from .assembly import description as asm_desc
from .assembly import visitor as asm_visitor
from .binning import description as binning_desc
from .binning import visitor as binning_visitor
from .classification import description as class_desc
from .classification import visitor as class_visitor


@final
class Topics(core.Topics):
    """Topic names."""

    ASSEMBLY = asm_desc.DESCRIPTION.name()
    CLASSIFICATION = class_desc.DESCRIPTION.name()
    BINNING = binning_desc.DESCRIPTION.name()

    def to_description(self) -> core.Description:
        """Get topic description."""
        match self:
            case Topics.ASSEMBLY:
                return asm_desc.DESCRIPTION
            case Topics.CLASSIFICATION:
                return class_desc.DESCRIPTION
            case Topics.BINNING:
                return binning_desc.DESCRIPTION

    def tools(self) -> type[core.Tools]:
        """Get topic tools."""
        match self:
            case Topics.ASSEMBLY:
                return asm_visitor.Tools
            case Topics.CLASSIFICATION:
                return class_visitor.Tools
            case Topics.BINNING:
                return binning_visitor.Tools



"""Binning topic visitor."""

from typing import final

from slurmbench.prelude.topic import visitor as core

from .pangebin_once import description as pgbonce_desc


@final
class Tools(core.Tools):
    """Plasmidness tools descriptions."""

    PANGEBIN_ONCE = pgbonce_desc.DESCRIPTION.name()

    def to_description(self) -> core.Description:
        """Get tool description."""
        match self:
            case Tools.PANGEBIN_ONCE:
                return pgbonce_desc.DESCRIPTION



"""Root pbfbench application module."""

# Due to typer usage:

from __future__ import annotations

from slurmbench.prelude import app as core_app

from .topics import visitor as topics_visitor
from .topics.assembly import app as assembly_app
from .topics.binning import app as binning_app
from .topics.classification import app as class_app

APP = core_app.new(
    "pbfbench",
    "PlasBin-flow benchmarking framework",
    topics_visitor.Topics,
    (
        assembly_app.APP,
        class_app.APP,
        binning_app.APP,
    ),
)



"""Concrete tool Bash script logics."""

from collections.abc import Iterator
from pathlib import Path
from typing import final

from slurmbench.prelude.tool import bash as core

import benchmark.topics.assembly.results as asm_res
from benchmark.topics.binning.plasbin_flow.format.classification import (
    results as fmt_class_res,
)


@final
class GFAInputLinesBuilder(core.Argument[asm_res.AsmGraphGZ]):
    """GFA input bash lines builder."""

    GFA_GZ_VAR = core.BashVar("GFA")

    def __gfa_gz_file(self) -> Path:
        """Return a gzipped GFA path with sample name is a sh variable."""
        return self._input_result.gfa_gz(
            self._input_data_smp_sh_fs_manager.sample_dir().name,
        )

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield self.GFA_GZ_VAR.set_path(self.__gfa_gz_file())


@final
class SeedsInputLinesBuilder(core.Argument[fmt_class_res.Seeds]):
    """Seeds input bash lines builder."""

    SEEDS_VAR = core.BashVar("SEEDS")

    def __plasbin_seeds_file(self) -> Path:
        """Return a gzipped GFA path with sample name is a sh variable."""
        return self._input_result.tsv(
            self._input_data_smp_sh_fs_manager.sample_dir().name,
        )

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield self.SEEDS_VAR.set_path(self.__plasbin_seeds_file())

    def close_lines(self) -> Iterator[str]:
        """Get shell input close lines."""
        yield from ()


@final
class PlasmidnessInputLinesBuilder(core.Argument[fmt_class_res.Plasmidness]):
    """Plasmidness input bash lines builder."""

    PLASMIDNESS_VAR = core.BashVar("PLASMIDNESS")

    def __plasbin_plasmidness_file(self) -> Path:
        """Return a gzipped GFA path with sample name is a sh variable."""
        return self._input_result.tsv(
            self._input_data_smp_sh_fs_manager.sample_dir().name,
        )

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield self.PLASMIDNESS_VAR.set_path(self.__plasbin_plasmidness_file())



"""plASgraph2 results."""

from __future__ import annotations

from pathlib import Path
from typing import final

from slurmbench.prelude.tool import results as core


@final
class PlasmidProbabilities(core.Original):
    """Plasmid probabilities result."""

    CSV_NAME = Path("plasmid_probabilities.csv")

    def csv(self, sample_dirname: str | Path) -> Path:
        """Get plasmid probabilities CSV file."""
        return self._exp_fs_manager.sample_dir(sample_dirname) / self.CSV_NAME



"""PlasBin-flow result formatting module."""

import pandas as pd
from slurmbench.prelude.tool import results as core

import benchmark.topics.assembly.visitor as asm_visitor
import benchmark.topics.classification.plasgraph2.results as plasgraph2_res
from benchmark.topics.classification.plasgraph2 import connector

from . import ops, results


def plasmidness(
    plasgraph2_data_exp_fs_manager: core.ExpFSDataManager,
    sample_item: core.Sample,
) -> results.Plasmidness:
    """Convert Platon result into plasmidness PlasBin-flow input."""
    plm_prob_res = plasgraph2_res.PlasmidProbabilities(
        plasgraph2_data_exp_fs_manager,
    )
    pbf_plasmidness_res = results.Plasmidness(
        plasgraph2_data_exp_fs_manager,
    )

    plasgraph2_gfa_arg = core.get_arg(plasgraph2_data_exp_fs_manager, connector.GFAArg)
    if isinstance(plasgraph2_gfa_arg, core.InvalidToolNameError):
        _err_msg = (
            f"Invalid tool name `{plasgraph2_gfa_arg.invalid_tool_name()}`"
            f" for argument name `{plasgraph2_gfa_arg.arg_name()}`."
        )
        raise TypeError(_err_msg)

    asm_data_fs_manager = core.ExpFSDataManager(
        plasgraph2_data_exp_fs_manager.root_dir(),
        asm_visitor.Tools(plasgraph2_gfa_arg.tool()).to_description(),
        plasgraph2_gfa_arg.exp_name(),
    )
    gfa_gz = (
        plasgraph2_gfa_arg.result_visitor()
        .result_builder()(asm_data_fs_manager)
        .gfa_gz(sample_item.uid())
    )

    plgr_df = pd.read_csv(plm_prob_res.csv(sample_item.uid()))
    contigs_dict = ops.parse_gfa(gfa_gz, plasgraph2_gfa_arg.tool())
    for _, row in plgr_df.iterrows():
        contig_id = str(row["contig"]).split(" ")[0]
        prcr, prpl = row["chrom_score"], row["plasmid_score"]
        pred = row["label"]
        if contig_id in contigs_dict:
            contigs_dict[contig_id]["Prob_Chromosome"] = float(prcr)
            contigs_dict[contig_id]["Prob_Plasmid"] = float(prpl)
            if pred == "plasmid":
                contigs_dict[contig_id]["Prediction"] = "Plasmid"
            else:
                contigs_dict[contig_id]["Prediction"] = "Chromosome"
    contigs_df = pd.DataFrame.from_dict(contigs_dict).T

    contigs_df.to_csv(
        pbf_plasmidness_res.tsv(sample_item.uid()),
        columns=["Prob_Plasmid"],
        sep="\t",
        header=False,
        index=True,
    )

    return pbf_plasmidness_res


def seeds(
    plasgraph2_data_exp_fs_manager: core.ExpFSDataManager,
    sample_item: core.Sample,
) -> results.Seeds:
    """Convert plasmid stats to PBF format."""
    plm_prob_res = plasgraph2_res.PlasmidProbabilities(
        plasgraph2_data_exp_fs_manager,
    )
    pbf_seeds_res = results.Seeds(
        plasgraph2_data_exp_fs_manager,
    )

    plasgraph2_gfa_arg = core.get_arg(plasgraph2_data_exp_fs_manager, connector.GFAArg)
    if isinstance(plasgraph2_gfa_arg, core.InvalidToolNameError):
        _err_msg = (
            f"Invalid tool name `{plasgraph2_gfa_arg.invalid_tool_name()}`"
            f" for argument name `{plasgraph2_gfa_arg.arg_name()}`."
        )
        raise TypeError(_err_msg)

    asm_data_fs_manager = core.ExpFSDataManager(
        plasgraph2_data_exp_fs_manager.root_dir(),
        asm_visitor.Tools(plasgraph2_gfa_arg.tool()).to_description(),
        plasgraph2_gfa_arg.exp_name(),
    )
    gfa_gz = (
        plasgraph2_gfa_arg.result_visitor()
        .result_builder()(asm_data_fs_manager)
        .gfa_gz(sample_item.uid())
    )

    seed_contigs: list[str] = []

    plgr_df = pd.read_csv(plm_prob_res.csv(sample_item.uid()))
    contigs_dict = ops.parse_gfa(gfa_gz, plasgraph2_gfa_arg.tool())
    for _, row in plgr_df.iterrows():
        contig_id = str(row["contig"]).split(" ")[0]
        pred = row["label"]
        if contig_id in contigs_dict and pred == "plasmid":
            seed_contigs.append(contig_id)

    with pbf_seeds_res.tsv(sample_item.uid()).open("w") as f:
        for contig_id in seed_contigs:
            f.write(f"{contig_id}\n")

    return pbf_seeds_res



"""PlasBin-flow result formatting module."""

import gzip

import pandas as pd
from Bio import SeqIO
from slurmbench.prelude.tool import results as core

import benchmark.topics.assembly.visitor as asm_visitor
import benchmark.topics.classification.platon.results as platon_res
from benchmark.topics.classification.platon import connector

from . import results


def plasmidness(
    platon_data_exp_fs_manager: core.ExpFSDataManager,
    sample_item: core.Sample,
) -> results.Plasmidness:
    """Convert Platon result into plasmidness PlasBin-flow input."""
    plasmidness_res = platon_res.PlasmidStats(platon_data_exp_fs_manager)
    pbf_plasmidness_res = results.Plasmidness(
        platon_data_exp_fs_manager,
    )

    platon_genome_arg = core.get_arg(
        platon_data_exp_fs_manager,
        connector.GenomeArg,
    )
    if isinstance(platon_genome_arg, core.InvalidToolNameError):
        _err_msg = (
            f"Invalid tool name `{platon_genome_arg.invalid_tool_name()}`"
            f" for argument name `{platon_genome_arg.arg_name()}`."
        )
        raise TypeError(_err_msg)

    asm_data_fs_manager = core.ExpFSDataManager(
        platon_data_exp_fs_manager.root_dir(),
        asm_visitor.Tools(platon_genome_arg.tool()).to_description(),
        platon_genome_arg.exp_name(),
    )
    fasta_gz = (
        platon_genome_arg.result_visitor()
        .result_builder()(asm_data_fs_manager)
        .fasta_gz(sample_item.uid())
    )

    with plasmidness_res.tsv(sample_item.uid()).open() as tsv_file:
        set_of_platon_ids = {line.split("\t")[0] for line in tsv_file}

    with (
        pbf_plasmidness_res.tsv(sample_item.uid()).open("w") as tsv_file,
        gzip.open(fasta_gz, "rt") as fasta_gz_file,
    ):
        for record in SeqIO.parse(fasta_gz_file, "fasta"):
            if record.name in set_of_platon_ids:
                tsv_file.write(f"{record.name}\t1\n")
            else:
                tsv_file.write(f"{record.name}\t0\n")

    return pbf_plasmidness_res


def seeds(
    platon_data_exp_fs_manager: core.ExpFSDataManager,
    sample_item: core.Sample,
) -> results.Seeds:
    """Convert plasmid stats to PBF format."""
    seeds_res = platon_res.PlasmidStats(platon_data_exp_fs_manager)
    pbf_seeds_res = results.Seeds(
        platon_data_exp_fs_manager,
    )

    platon_seeds_stats_df = pd.read_csv(
        seeds_res.tsv(sample_item.uid()),
        sep="\t",
    )

    platon_seeds_stats_df.to_csv(
        pbf_seeds_res.tsv(sample_item.uid()),
        columns=["ID"],
        header=False,
        sep="\t",
        index=False,
    )
    return pbf_seeds_res
